from collections import OrderedDict

import numpy as np
import torch
import torch.optim as optim
from torch import nn as nn

import rlkit.torch.pytorch_util as ptu
from rlkit.core.eval_util import create_stats_ordered_dict
from rlkit.torch.torch_rl_algorithm import TorchTrainer


class FullStepSimpleTrainer(TorchTrainer):
    """
    Trainer for Behavior Cloning
    Policy is trained by maximizing log likelihood of actions in a given dataset.
    Q function is trained by SARSA
    """
    def __init__(
            self,
            env,
            policy,
            target_policy,
            policy_data,
            qf1,
            target_qf1,

            kl_reg=True,
            is_q_type='none',
            log_is_clip_value=1.0,
            n_actions=10,
            alpha=1.0,
            discount=0.99,
            reward_scale=1.0,

            n_train_steps_qf_per_train_loop=1,
            qf_lr=1e-4,
            policy_lr=1e-4,
            optimizer_class=optim.Adam,

            soft_target_tau=5e-3,
            target_update_period=2,
            update_policy_data=False,
            normalize_q=False,
    ):
        super().__init__()
        self.env = env
        self.policy = policy
        self.target_policy = target_policy
        self.policy_data = policy_data
        self.qf1 = qf1
        self.target_qf1 = target_qf1
        self.soft_target_tau = soft_target_tau
        self.target_update_period = target_update_period

        self.update_policy_data = update_policy_data
        print('self.update_policy_data: \t', self.update_policy_data)
        self.kl_reg = kl_reg
        print('self.kl_reg: \t', self.kl_reg)
        self.normalize_q = normalize_q
        print('self.normalize_q: \t', self.normalize_q)

        self.is_q_type = is_q_type
        print('self.is_q_type: \t', self.is_q_type)
        self.log_is_clip_value = log_is_clip_value
        if self.is_q_type != 'none':
            print("self.log_is_clip_value: \t", self.log_is_clip_value)

        self.n_actions = n_actions
        self.alpha = alpha
        print('self.alpha: \t', self.alpha)

        self.qf_criterion = nn.MSELoss()

        self.policy_lr = policy_lr
        self.qf_lr = qf_lr
        self.optimizer_class = optimizer_class

        self.policy_optimizer = self.optimizer_class(
            self.policy.parameters(),
            lr=self.policy_lr,
        )
        self.qf1_optimizer = self.optimizer_class(
            self.qf1.parameters(),
            lr=self.qf_lr,
        )

        self.discount = discount
        self.reward_scale = reward_scale
        self.eval_statistics = OrderedDict()
        self.eval_wandb = OrderedDict()
        self._n_train_steps_total = 0

        self._n_train_steps_qf_per_train_loop = n_train_steps_qf_per_train_loop
        print('self._n_train_steps_qf_per_train_loop: \t', self._n_train_steps_qf_per_train_loop)
        self._need_to_update_eval_statistics = True

        self.discrete = False

    def train_from_torch(self, batch):
        rewards = batch['rewards']
        terminals = batch['terminals']
        obs = batch['observations']
        actions = batch['actions']
        next_obs = batch['next_observations']

        # Update Q first
        """
        QF Loss
        """
        q1_pred = self.qf1(obs, actions)

        if self.is_q_type in ['simple', 'self-norm']:
            next_actions, _, _, log_prob_data, *_ = self.policy_data(
                next_obs, reparameterize=False, return_log_prob=True,
            )
            log_prob_pi = self.target_policy.log_prob(next_obs, next_actions).unsqueeze(-1)

            importance_sampling = torch.exp(
                torch.clip(log_prob_pi - log_prob_data, -1 * self.log_is_clip_value, self.log_is_clip_value))
            if self.is_q_type == 'self-norm':
                importance_sampling = importance_sampling / importance_sampling.mean()
        else:
            next_actions, *_ = self.target_policy(
                next_obs, reparameterize=False, return_log_prob=False,
            )
            importance_sampling = ptu.ones_like(q1_pred)

        # target_qf1_values = self.target_qf1(next_obs, next_actions)
        # target_qf2_values = self.target_qf2(next_obs, next_actions)
        # target_q_values = torch.min(target_qf1_values, target_qf2_values)
        target_q_values = self.target_qf1(next_obs, next_actions)
        q_target = self.reward_scale * rewards + (1. - terminals) * self.discount * importance_sampling * target_q_values

        qf1_loss = self.qf_criterion(q1_pred, q_target.detach())
        # qf2_loss = self.qf_criterion(q2_pred, q_target.detach())
        qf2_loss = torch.tensor(0.)
        qf_loss = qf1_loss + qf2_loss

        self.qf1_optimizer.zero_grad()
        qf1_loss.backward()
        self.qf1_optimizer.step()

        # self.qf2_optimizer.zero_grad()
        # qf2_loss.backward()
        # self.qf2_optimizer.step()

        # Policy is updated every self._n_train_steps_qf_per_train_loop steps
        if self._n_train_steps_total % self._n_train_steps_qf_per_train_loop == 0:
            """
            Policy and Alpha Loss
            """
            new_obs_actions, policy_mean, policy_log_std, log_pi, *_ = self.policy(
                obs, reparameterize=True, return_log_prob=True,
            )

            if self.kl_reg:
                obs_stack = torch.unsqueeze(obs, 1).repeat(1, self.n_actions, 1).reshape((-1, obs.shape[1]))
                new_obs_actions_stack, _, _, log_pi_stack, *_ = self.policy(obs_stack, reparameterize=True, return_log_prob=True, )
                log_pi = torch.mean(log_pi_stack.reshape((-1, self.n_actions)), dim=1)

                log_pi_data_stack = self.policy_data.log_prob(obs_stack, new_obs_actions_stack)
                log_pi_data = torch.mean(log_pi_data_stack.reshape((-1, self.n_actions)), dim=1)

                kl = (log_pi - log_pi_data).mean()
                qf1_pi = self.qf1(obs, new_obs_actions)

                if self.normalize_q:
                    avg_qf = qf1_pi.abs().mean().detach()
                    policy_loss = self.alpha * kl - qf1_pi.mean() / avg_qf
                else:
                    policy_loss = self.alpha * kl - qf1_pi.mean()
            else:
                policy_loss = -1 * self.qf1(obs, new_obs_actions).mean()
                kl = torch.tensor(0.)

            self.policy_optimizer.zero_grad()
            policy_loss.backward()
            self.policy_optimizer.step()


            """
            Soft Updates
            """
            if self._n_train_steps_total % (self._n_train_steps_qf_per_train_loop * self.target_update_period) == 0:
                ptu.soft_update_from_to(
                    self.policy, self.target_policy, self.soft_target_tau
                )
        else:
            policy_loss = kl = torch.tensor(0.)

        """
        Save some statistics for eval
        """
        if self._need_to_update_eval_statistics:
            self._need_to_update_eval_statistics = False
            """
            Eval should set this to None.
            This way, these statistics are only computed for one batch.
            """

            self.eval_statistics['Policy Loss'] = np.mean(ptu.get_numpy(policy_loss))
            self.eval_statistics['QF1 Loss'] = np.mean(ptu.get_numpy(qf1_loss))
            self.eval_statistics['QF2 Loss'] = self.eval_statistics['QF1 Loss']
            self.eval_statistics['QF Loss'] = np.mean(ptu.get_numpy(qf_loss))
            self.eval_statistics.update(create_stats_ordered_dict(
                'Q1 Predictions',
                ptu.get_numpy(q1_pred),
            ))
            # self.eval_statistics.update(create_stats_ordered_dict(
            #     'Q2 Predictions',
            #     ptu.get_numpy(q2_pred),
            # ))

            if self.kl_reg:
                self.eval_statistics['KL'] = np.mean(ptu.get_numpy(kl))

            for k in ['Policy Loss', 'Q1 Predictions', 'Q2 Predictions', 'QF1 Loss', 'QF2 Loss', 'QF Loss', 'KL']:
                self.eval_wandb[k] = self.eval_statistics.get(k, 0)

            for k1 in ['Q1 Predictions', 'Q2 Predictions']:
                for k2 in ['Mean', 'Std', 'Max', 'Min']:
                    k = ' '.join([k1, k2])
                    self.eval_wandb[k] = self.eval_statistics.get(k, 0)

        self._n_train_steps_total += 1

    def get_diagnostics(self):
        return (self.eval_statistics, self.eval_wandb)

    def end_epoch(self, epoch):
        self._need_to_update_eval_statistics = True

    @property
    def networks(self):
        return [
            self.policy,
            self.target_policy,
            self.policy_data,
            self.qf1,
            self.target_qf1,
        ]

    def get_snapshot(self):
        return dict(
            policy=self.policy,
            policy_data=self.policy_data,
            qf1=self.qf1,
            target_qf1=self.target_qf1,
        )

    def set_snapshot(self, snapshot):
        self.policy = snapshot['policy']
        self.policy_data = snapshot['policy_data']
        self.qf1 = snapshot['qf1']
        self.target_qf1 = snapshot['target_qf1']
